module("L_HoneywellLib_lyric", package.seeall)

--------------------------------------------------------------------------------
--
-- Honeywell Total Connect Comfort backend for Lyric API. Finally Honeywell has
-- released a public API and this backend is derived from it. The app is
-- registered with Honeywell and assigned a client ID/secret for authentication
--
-- Creator:  Mikee

-- Version 1.0: Dec 27, 2015 Mikee
--   Initial cut. Working authentication. Will need a device to develop further.

local main = { }

-- Upcalls
local trace = nil
local log = nil
local log_http = nil
local err = nil
local dump = nil
local url_encode = nil
local inject = nil
local trim = nil
local task = nil
local clearTask = nil
local lua_to_url = nil
local updateCookies = nil
local createCookie = nil
local xml_to_lua = nil

local json = require("L_HoneywellTCC_json")
local https = require("ssl.https")

--------------------------------------------------------------------------------
-- Constants

-- various constants
local TCC_SID = "urn:joeyd-com:serviceId:HoneywellTCC1"
local PROP_REAUTH_INTERVAL = "ReAuthInterval"

-- Status codes
local STATUS_OK = 1
local STATUS_FAIL = 2
local STATUS_REMOTE_FAIL = 3
local STATUS_AUTH_FAIL = 4
local STATUS_HTTP_FAIL = 5

-- App constants
local K_HOST = "api.honeywell.com"
local K_APP_KEY = "Z487ZykgPp3Su03tGtXzEwuvOdf2IYFb"
local K_APP_SECRET = "4Vg8JGJXyxpQhcDC"

local RetryLimit = 3

--------------------------------------------------------------------------------
-- Globals

local Cookies = { }
local RefreshToken = ""
local AccessToken = nil
local AccessTokenExpires = 0
local RefreshTime = 0
local TccDevice = 0

--------------------------------------------------------------------------------
-- Utility functions

-- character table string
local b='ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'

-- base64 encoding
function base64_enc(data)
   return ((data:gsub('.', function(x) 
			 local r,b='',x:byte()
			 for i=8,1,-1 do r=r..(b%2^i-b%2^(i-1)>0 and '1' or '0') end
        return r;
			   end)..'0000'):gsub('%d%d%d?%d?%d?%d?', function(x)
						 if (#x < 6) then return '' end
        local c=0
        for i=1,6 do c=c+(x:sub(i,i)=='1' and 2^(6-i) or 0) end
        return b:sub(c+1,c+1)
								  end)..({ '', '==', '=' })[#data%3+1])
end

-- base64 decoding
function base64_dec(data)
   data = string.gsub(data, '[^'..b..'=]', '')
   return (data:gsub('.', function(x)
			if (x == '=') then return '' end
			local r,f='',(b:find(x)-1)
			for i=6,1,-1 do r=r..(f%2^i-f%2^(i-1)>0 and '1' or '0') end
        return r;
			  end):gsub('%d%d%d?%d?%d?%d?%d?%d?', function(x)
				       if (#x ~= 8) then return '' end
        local c=0
        for i=1,8 do c=c+(x:sub(i,i)=='1' and 2^(8-i) or 0) end
        return string.char(c)
							      end))
end

--++
-- Make an http request with retry
--
-- @param spec_options - table of additional or override options
-- @return res - 
--         code - https result code
--         header - response headers
--
function https_request(spec_options)
   trace("https_request", { url = spec_options.url })

   local res, code, header 
   local base_options = {
      ["headers"] = {
	 ["Accept"] = "*/*",
--	 ["Accept-Encoding"] = "gzip,deflate,sdch",
	 ["Accept-Encoding"] = "sdch",
	 ["Accept-Language"] = "en-US",
	 ["Cache-Control"] = "max-age=0",
	 ["Connection"] = "keep-alive",
	 ["Host"] = K_HOST,
      },
   }

   for i=1,RetryLimit do
      local options = { }
      options.headers = { }
      for name,value in pairs(base_options) do
	 if name == "headers" then
	    for hname, hvalue in pairs(value) do
	       options.headers[hname] = hvalue
	    end
	 else
	    options[name] = value
	 end
      end
      for name,value in pairs(spec_options) do
	 if name == "headers" then
	    for hname, hvalue in pairs(value) do
	       options.headers[hname] = hvalue
	    end
	 else
	    options[name] = value
	 end
      end

      local cookie = createCookie(Cookies)
      if cookie ~= "" then
	 options.headers["Cookie"] = cookie
      end

      log_http("https req", { options=options })
      res, code, header = https.request(options)
      log_http("https resp", { res=res, code=code, header=header })

      if code ~= 400 and code ~= 504 then
	 break
      end
      trace("https_request retry", { res=res, code=code })
   end

   trace("https_request return", { res=res, code=code })
   return res, code, header
end

local function find_device(devices, deviceID)
   trace("find_device") --, { devices=devices, deviceID=deviceID })
   for _,dev in pairs(devices) do
      if dev.deviceID == deviceID then
	 trace("find_device return dev") --, { dev=dev })
	 return dev
      end
   end
   trace("find_device return")
   return nil
end

--------------------------------------------------------------------------------
-- Support functions

--++
-- Refresh the login token
--
local function refreshLogin()
   trace("refreshLogin")

   local status = STATUS_FAIL
   local reason = "Unknown failure"

   local result = { }
   local body = {
      grant_type="refresh_token",
      redirect_uri="oob",
      client_id=K_APP_KEY,
      client_secret=K_APP_SECRET,
      refresh_token=RefreshToken,
   }
   body = lua_to_url(body)
   log_http("body", { body=body })

   options = {
      ["redirect"] = false,
      ["method"] = "POST",
      ["url"] = "https://" .. K_HOST .. "/oauth2/token",
      ["headers"] = {
	 ["Authorization"] = base64_enc(K_APP_KEY .. ':' .. K_APP_SECRET),
	 ["Accept"] = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
	 ["Accept-Encoding"] = "sdch",
	 ["Content-Type"] = "application/x-www-form-urlencoded",
	 ["Content-Length"] = #body,
      },
      ["source"] = ltn12.source.string(body),
      ["sink"] = ltn12.sink.table(result),
   }

   log_http("token req", { options=options })
   res, code, header = https_request(options)
   log_http("token resp", { res=res, code=code, header=header, result=result })

   if code == 200 then
      status = STATUS_OK
      reason = "Success"

      res = json.decode(table.concat(result))
      log("access res", { res=res })
      AccessToken = res.access_token
      RefreshToken = res.refresh_token
      RefreshTime = tonumber(res.expires_in)
      AccessTokenExpires = os.time() + RefreshTime

      luup.variable_set(TCC_SID, PROP_REAUTH_INTERVAL, RefreshTime - 30, TccDevice)
   end

   trace("refreshLogin return", { status=status, reason=reason })
   return status, reason
end

local function getStatus(locationID, deviceID)
   trace("getStatus", { locationID=locationID, deviceID=deviceID })

   local status = STATUS_FAIL
   local reason = "Unknown failure"

   local result = { }
   local options = {
      ["method"] = "GET",
      ["url"] = "https://" .. K_HOST .. "/v1/devices?apikey=" .. K_APP_KEY ..
	 "&locationId=" .. locationID,
      ["headers"] = {
	 ["Authorization"] = "Bearer " .. AccessToken,
	 ["Content-Type"] = "application/json",
      },
      ["sink"] = ltn12.sink.table(result),
   }
   log_http("thermostat req", { options=options })
   local res, code, header = https_request(options)
   log_http("thermostat resp", { res=res, code=code, header=header, result=result })

   local dev
   if code == 200 then

      log_http("thermostat json =", table.concat(result))
      result = json.decode(table.concat(result))

      if result.fault == nil then
	 dev = find_device(result, deviceID)
	 if dev ~= nil then
	    status = STATUS_OK
	    reason = "Success"
	 end
      end
   elseif code == 401 then
      status = STATUS_AUTH_FAIL
      reason = "Authentication failed"
   end

   trace("getStatus return", { status=status, reason=reason }) --, dev=dev })
   return status, reason, dev
end

--------------------------------------------------------------------------------
-- APIs

function Init(devId, imports)
   TccDevice = devId
   main = imports
   trace = main.trace
   log = main.log
   log_http = main.log_http
   err = main.err

   dump = main.dump

   inject = main.inject
   trim = main.trim

   task = main.task
   clearTask = main.clearTask

   lua_to_url = main.lua_to_url
   xml_to_lua = main.xml_to_lua
   url_encode = main.url_encode

   updateCookies = main.updateCookies
   createCookie = main.createCookie

end

--++
-- Initiate an OAuth2 3 leg authentication. This runs as follows:
-- 1. Send a request to the server (honeywell) with our client id (registered
--    with honeywell)
-- 2. Get a redirect to the honeywell login page
-- 3. Issue a POST with the honeywell credentials
-- 4. Get a redirect with the consent to allow the app to use the credentials
--    we just supplied
-- 5. Issue a POST agreeing to use the creds
-- 6. Get a redirect to our URL ("oob" as registered with honeywell) with the
--    scope and code as uri search params, i.e. iri?scope=&code=xxx.
-- 7. Make a request of the server (honeywell) for the access token given the
--    code above.
--
-- @param user - string user
-- @param pass - string pass
-- @return status - STATUS_OK, STATUS_xxx
--         reason - reason for status above
--         token - token to use on subsequent calls
--
function Login(user, pass)
   trace("Lyric.Login")

   local status = STATUS_FAIL
   local reason = "Unknown failure"

   if AccessTokenExpires > os.time() then
      status, reason = refreshLogin()
      if status == STATUS_OK then
	 trace("Lyric.Login return (refresh)", { status=status, reason=reason, token=AccessToken })
	 return status, reason, AccessToken
      end
      AccessToken = nil
      -- fall thru
   end

   local result = { }
   local options = {
      ["redirect"] = false,
      ["method"] = "GET",
      ["url"] = "https://" .. K_HOST .. "/oauth2/authorize?client_id=" .. K_APP_KEY .. "&redirect_uri=oob&response_type=code",
      ["headers"] = {
	 ["Content-Type"] = "application/x-www-form-urlencoded;charset=utf-8",
      },
      ["sink"] = ltn12.sink.table(result),
   }
   log_http("authorize req", { options=options })
   local res, code, header = https_request(options)
   log_http("authorize resp", { res=res, code=code, header=header, result=result })

   -- should redirect to login
   if code == 302 then
      result = { }
      Cookies = updateCookies(header["set-cookie"], Cookies)

      url = header.location
      url = string.gsub(url, "%s", "%%20")
      log_http("url", { url=url })
      local body = {
	 username=user,
	 password=pass,
	 submit="Login"
      }
      body = lua_to_url(body)
      log_http("body", { body=body })

      options = {
	 ["method"] = "POST",
	 ["url"] = url,
	 ["headers"] = {
	    ["Accept"] = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
	    ["Accept-Encoding"] = "sdch",
	    ["Content-Type"] = "application/x-www-form-urlencoded",
	    ["Content-Length"] = #body,
	    ["Referer"] = url,
	 },
	 ["source"] = ltn12.source.string(body),
	 ["sink"] = ltn12.sink.table(result),
      }
      log_http("login req", { options=options })
      res, code, header = https_request(options)
      log_http("login resp", { res=res, code=code, header=header, result=result })

      -- should redirect to consent
      if code == 302 then
	 result = { }
	 Cookies = updateCookies(header["set-cookie"], Cookies)

	 url = header.location
	 url = string.gsub(url, "%s", "%%20")
	 log_http("url", { url=url })
	 local body = {
	    decision="yes",
	 }
	 body = lua_to_url(body)
	 log_http("body", { body=body })

	 options = {
	    ["method"] = "POST",
	    ["url"] = "https://" .. K_HOST .. url,
	    ["headers"] = {
	       ["Accept"] = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
	       ["Accept-Encoding"] = "sdch",
	       ["Content-Type"] = "application/x-www-form-urlencoded",
	       ["Content-Length"] = #body,
	    },
	    ["source"] = ltn12.source.string(body),
	    ["sink"] = ltn12.sink.table(result),
	 }

	 log_http("consent req", { options=options })
	 res, code, header = https_request(options)
	 log_http("consent resp", { res=res, code=code, header=header, result=result })

	 -- should redirect to oob (the URL registered with "Vera plugin")
	 if code == 302 then
	    result = { }
	    log_http("location", { loc=header.location })
	    local uri, params = string.match(header.location, "([^%?]-)%?(.*)")
	    log_http("parsed url", { uri=uri, params=params })

	    if uri == "oob" then

	       local auth_code = ""
	       for param in string.gmatch(params,"([^%&]*)%&?") do
		  for name,value in string.gmatch(param,"([^=]-)=(.*)") do
		     if name == "code" then
			auth_code = value
		     end
		  end
	       end

	       if (auth_code ~= "") then

		  local body = {
		     grant_type="authorization_code",
		     code=auth_code,
		     redirect_uri="oob",
		     client_id=K_APP_KEY,
		     client_secret=K_APP_SECRET,
		  }
		  body = lua_to_url(body)
		  log_http("body", { body=body })

		  options = {
		     ["redirect"] = false,
		     ["method"] = "POST",
		     ["url"] = "https://" .. K_HOST .. "/oauth2/token",
		     ["headers"] = {
			["Authorization"] = base64_enc(K_APP_KEY .. ':' .. K_APP_SECRET),
			["Accept"] = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
			["Accept-Encoding"] = "sdch",
			["Content-Type"] = "application/x-www-form-urlencoded",
			["Content-Length"] = #body,
		     },
		     ["source"] = ltn12.source.string(body),
		     ["sink"] = ltn12.sink.table(result),
		  }

		  log_http("token req", { options=options })
		  res, code, header = https_request(options)
		  log_http("token resp", { res=res, code=code, header=header, result=result })

		  if code == 200 then
		     status = STATUS_OK
		     reason = "Success"

		     res = xml_to_lua(table.concat(result))
		     log_http("access res", { res=res })
		     AccessToken = res.Root.access_token
		     RefreshToken = res.Root.refresh_token
		     RefreshTime = tonumber(res.Root.expires_in)
		     AccessTokenExpires = os.time() + RefreshTime

		     log("Schedule login refresh in " .. tostring(RefreshTime - 30) .. " secs")
		     luup.variable_set(TCC_SID, PROP_REAUTH_INTERVAL, RefreshTime - 30, TccDevice)
		  else
		     status = STATUS_FAIL
		     reason = "Failed to obtain access token"
		  end

	       end
	    else
	       status = STATUS_FAIL
	       reason = "No redirection back to us"
	    end
	 end
      else
	 status = STATUS_FAIL
	 reason = "No redirection"
      end
   else
      status = STATUS_FAIL
      reason = "No initial redirection"
   end

   trace("Lyric.Login return", { status=status, reason=reason, token=AccessToken })
   return status, reason, AccessToken
end

--++
-- Return a structure of location/device information
--
-- @param token - authentication token
-- @return status - STATUS_OK, STATUS_xxx
--         reason - reason for status above
--         devices {
--                  { locationID = {
--                    Name = "",  - string location name
--                    Devices = {
--                     { deviceId = name },...
--                    }
--                   },...
--                  }
function GetDevices(token)
   trace("Lyric.GetDevices", { token=token })

   AccessToken = token
   if AccessToken == nil then
      err("Invalid token")
      return STATUS_AUTH_FAIL, "Invalid token", nil
   end

   local locations = { }
   local status = STATUS_FAIL
   local reason = "Unknown failure"

   local result = { }
   local options = {
      ["method"] = "GET",
      ["url"] = "https://" .. K_HOST .. "/v1/locations?apikey=" .. K_APP_KEY,
      ["headers"] = {
	 ["Authorization"] = "Bearer " .. AccessToken,
	 ["Content-Type"] = "application/xml",
      },
      ["sink"] = ltn12.sink.table(result),
   }
   log_http("location req", { options=options })
   local res, code, header = https_request(options)
   log_http("location resp", { res=res, code=code, header=header, result=result })

   if code == 200 then
-- test
--      local f = io.open("lyric_locations.json", "r");
--      result = f:read("*all")
--      f:close()
-- end test

      log_http("location json =", table.concat(result))
      result = json.decode(table.concat(result))
      for _,loc in ipairs(result) do
	 tbl = { }
	 for _,dev in ipairs(loc.devices) do
	    tbl[loc.locationID .. ":" .. dev.deviceID] = { ["Name"] = dev.name }
	 end
	 locations[loc.locationID] = { }
	 locations[loc.locationID].Name = loc.name
	 locations[loc.locationID].Devices = tbl
      end

      status = STATUS_OK
      reason = "Success"
   elseif code == 401 then
      status = STATUS_AUTH_FAIL
      reason = "Authentication failed"
   else
      status = STATUS_FAIL
      reason = "Failed to get devices"
   end

   trace("Lyric.GetDevices return", { status=status, reason=reason, locations=locations })
   return status, reason, locations
end

--++
-- Return the status of a thermostat
--
-- @param token - authentication token
-- @param deviceId - device id
-- @return status - STATUS_OK, STATUS_xxx
--         reason - reason for status above
--         status {
--		Success,
--		Load,
--		FanSwitchPosition,
--		HasFan,
--		CoolNextPeriod,
--		CoolSetpoint,
--		DisplayUnits,
--		DisplayTemp,
--		HeatNextPeriod,
--		HeatSetpoint,
--		IndoorHumidity,
--		HasIndoorHumidity,
--		OutdoorHumidity,
--		HasOutdoorHumidity,
--		OutdoorTemp,
--		HasOutdoorTemp,
--		SchedCoolSetpoint,
--		SchedHeatSetpoint,
--		StatusCool,
--		StatusHeat,
--		CanSwitchAuto,
--		SystemSwitchPosition
--         }
function GetStatus(token, deviceId)
   trace("Lyric.GetStatus", { token=token, deviceId=deviceId })

   local locationID, deviceID = string.match(deviceId, "([^:]-):(.*)")

   AccessToken = token
   if AccessToken == nil then
      err("Invalid token")
      return STATUS_AUTH_FAIL, "Invalid token", nil
   end

   local status = STATUS_FAIL
   local reason = "Unknown failure"

   local status, reason, dev = getStatus(locationID, deviceID)

   local data = { }
   if status == STATUS_OK and dev ~= nil then
      status = STATUS_OK
      reason = "Success"

      data.Success = true
      data.Load = 0
      data.CoolNextPeriod = 0
      data.CoolSetpoint = dev.thermostat.changeableValues.coolSetpoint
      data.DisplayUnits = (dev.thermostat.units == "Fahrenheit" and "F" or "C")
      data.DisplayTemp = dev.thermostat.indoorTemperature
      data.HeatNextPeriod = 0
      data.HeatSetpoint = dev.thermostat.changeableValues.heatSetpoint
      data.IndoorHumidity = dev.thermostat.indoorHumidity
      data.HasIndoorHumidity = (dev.thermostat.indoorHumidityStatus == "Measured" and true or false)
      data.OutdoorHumidity = 0
      data.HasOutdoorHumidity = false
      data.OutdoorTemp = dev.thermostat.outdoorTemperature
      data.HasOutdoorTemp = true
      data.SchedCoolSetpoint = 0
      data.SchedHeatSetpoint = 0
      data.StatusCool = 0
      data.StatusHeat = 0
      data.CanSwitchAuto = true
      local sysmode = dev.thermostat.changeableValues.mode
      data.SystemSwitchPosition = 0
      if sysmode == "Off" then
	 data.SystemSwitchPosition = 2
      elseif sysmode == "Heat" then
	 data.SystemSwitchPosition = 1
      elseif sysmode == "Cool" then
	 data.SystemSwitchPosition = 3
      end

      local fanmode = dev.settings.fan.changeableValues.mode
      data.FanSwitchPosition = 0
      if fanmode == "Auto" then
	 data.FanSwitchPosition = 0
      elseif fanmode == "On" then
	 data.FanSwitchPosition = 1
      elseif fanmode == "Circulate" then
	 data.FanSwitchPosition = 2
      end
      data.HasFan = 0
   else
      status = STATUS_FAIL
      reason = "Failed to get status"
   end

   trace("Lyric.GetStatus return", { status=status, reason=reason, data=data })
   return status, reason, data
end

--++
-- Set the operating parameters of a thermostat
--
-- @param token - authentication token
-- @param deviceId - device id
-- @param req - table in the following format
--              { DeviceID = tonumber(device_id),
--                SystemSwitch = n,     -- 2=Off, 1=Heat, 3=Cool, 4=Auto
--                HeatSetpoint = n,     -- temperature
--                CoolSetpoint = n,     -- temperature
--                HeatNextPeriod = n,   -- integer representing a 15 minute period of day. 0=12:00 AM, 1=12:15, 4 = 1:00 AM etc.
--                CoolNextPeriod = n,   -- integer representing a 15 minute period of day. 0=12:00 AM, 1=12:15, 4 = 1:00 AM etc.
--                StatusHeat = n,       -- Set to 1 when setting a heat point "until next program", 2 for permanent hold, 0 for cancel
--                StatusCool = n,	   -- Set to 1 when setting a cool point "until next program", 2 for permanent hold, 0 for cancel
--                FanMode = n,          -- 2=Circulate, 0=Auto, 1=On, 3=Follow Shedule
--              }
-- @return status - STATUS_OK, STATUS_xxx
--         reason - reason for status above
--
function SetStatus(token, deviceId, req)
   trace("Lyric.SetStatus", { token=token, deviceId=deviceId })

   AccessToken = token
   if AccessToken == nil then
      err("Invalid token")
      return STATUS_AUTH_FAIL, "Invalid token", nil
   end

   local locationID, deviceID = string.match(deviceId, "([^:]-):(.*)")

   local status, reason, current = getStatus(locationID, deviceID)
   if status ~= STATUS_OK then
      return status, reason
   end

-- "mode": "Off",
--  "autoChangeoverActive": true,
--  "emergencyHeatActive": false, (fails)
--  "heatSetpoint": 72,
--  "coolSetpoint": 78

   status = STATUS_FAIL
   reason = "Unknown failure"

   local body = { }
   if req.SystemSwitch == nil then
      body.mode = current.thermostat.changeableValues.mode
      body.autoChangeoverActive = current.thermostat.changeableValues.autoChangeoverActive
   end
   if req.SystemSwitch == 2 then
      body.mode = "Off"
      body.autoChangeoverActive = false
   elseif req.SystemSwitch == 1 then
      body.mode = "Heat"
      body.autoChangeoverActive = false
   elseif req.SystemSwitch == 3 then
      body.mode = "Cool"
      body.autoChangeoverActive = false
   elseif req.SystemSwitch == 4 then
      body.autoChangeoverActive = true
   end

   body.heatSetpoint = tonumber(req.HeatSetpoint)
   body.coolSetpoint = tonumber(req.CoolSetpoint)

   if not current.thermostat.hasDualSetpointStatus then
      if body.mode == "Heat" then
	 body.coolSetpoint = body.heatSetpoint
      elseif body.mode == "Cool" then
	 body.heatSetpoint = body.coolSetpoint
      end
   end

   if body.heatSetpoint == nil then
      body.heatSetpoint = current.thermostat.changeableValues.heatSetpoint
   end
   if body.coolSetpoint == nil then
      body.coolSetpoint = current.thermostat.changeableValues.coolSetpoint
   end

   body = json.encode(body)
--   log_http("body", { body=body })
   log("body", { body=body })

   local result = { }
   local options = {
      ["method"] = "POST",
      ["url"] = "https://" .. K_HOST .. "/v1/devices/thermostats/" ..
	 deviceID .. "?apikey=" .. K_APP_KEY .. "&locationId=" ..
	 locationID,
      ["headers"] = {
	 ["Authorization"] = "Bearer " .. AccessToken,
	 ["Content-Type"] = "application/json",
	 ["Content-Length"] = #body,
      },
      ["source"] = ltn12.source.string(body),
      ["sink"] = ltn12.sink.table(result),
   }
   log_http("thermostat req", { options=options })
   local res, code, header = https_request(options)
   log_http("thermostat resp", { res=res, code=code, header=header, result=result })

   if code == 200 then
      status = STATUS_OK
      reason = "Success"

      body = { }
      if req.FanMode ~= nil then
	 if req.FanMode == 0 then
	    body.mode = "Auto"
	 elseif req.FanMode == 2 then
	    body.mode = "Circulate"
	 elseif req.FanMode == 1 then
	    body.mode = "On"
	 end
	 body = json.encode(body)
	 --      log_http("body", { body=body })
	 log("body", { body=body })

	 local options = {
	    ["method"] = "POST",
	    ["url"] = "https://" .. K_HOST .. "/v1/devices/thermostats/" ..
	       deviceID .. "/fan?apikey=" .. K_APP_KEY .. "&locationId=" ..
	       locationID,
	    ["headers"] = {
	       ["Authorization"] = "Bearer " .. AccessToken,
	       ["Content-Type"] = "application/json",
	       ["Content-Length"] = #body,
	    },
	    ["source"] = ltn12.source.string(body),
	    ["sink"] = ltn12.sink.table(result),
	 }
	 log_http("fan req", { options=options })
	 local res, code, header = https_request(options)
	 log_http("fan resp", { res=res, code=code, header=header })--, result=result })

	 if code == 200 then
	    -- empty
	 elseif code == 401 then
	    status = STATUS_AUTH_FAIL
	    reason = "Authentication failed"
	 else
	    status = STATUS_FAIL
	    reason = "Failed to set fan mode"
	 end
      end
   elseif code == 401 then
      status = STATUS_AUTH_FAIL
      reason = "Authentication failed"
   end

   trace("Lyric.SetStatus return", { status=status, reason=reason })
   return status, reason
end
